#include "cpu.h"

#include "instruction.h"
#include "counter.h"

#include "config.h"

#include <iostream>
#include <string>
#include <sstream>

#include <boost/tokenizer.hpp>
#include <boost/algorithm/string/trim.hpp>

using namespace std;
using namespace boost;

CCPU::CCPU() :
	m_instruction(),
	m_program(),
	m_label(),
	m_array(),
	m_asm_size(0)
{
	/** Asm instructions */
	m_instruction.insert(make_pair("mov.w", new CMov()));
	m_instruction.insert(make_pair("mov.b", new CMov()));

	m_instruction.insert(make_pair("add.w", new CAdd()));
	m_instruction.insert(make_pair("add.b", new CAdd()));
	m_instruction.insert(make_pair("adds", new CAddConstant()));

	m_instruction.insert(make_pair("sub.w", new CSub()));
	m_instruction.insert(make_pair("sub.b", new CSub()));
	m_instruction.insert(make_pair("subs", new CSubConstant()));

	m_instruction.insert(make_pair("cmp.w", new CCompare()));
	m_instruction.insert(make_pair("cmp.b", new CCompare()));
	m_instruction.insert(make_pair("rts", new CReturn()));

	m_instruction.insert(make_pair("bra", new CJump(m_label)));
	m_instruction.insert(make_pair("beq", new CJumpEqual(m_label)));
	m_instruction.insert(make_pair("bne", new CJumpNotEqual(m_label)));
	m_instruction.insert(make_pair("jsr", new CJumpToSubroutine(m_label)));

	/** Asm directives */
	m_instruction.insert(make_pair(".section", new SSection(m_label))); // jump to main
	m_instruction.insert(make_pair(".align", new SAlign()));
	m_instruction.insert(make_pair(".space", new SAlign()));
	m_instruction.insert(make_pair(".global", new SGlobal()));
	m_instruction.insert(make_pair(".end", new SEnd()));

	/** Meta directives */
	m_instruction.insert(make_pair(".file", new SFile())); // initial code
	m_instruction.insert(make_pair(".ident", new SIdent())); // final code

	/** Array directives */
	m_instruction.insert(make_pair(".ascii", new SAscii())); // final code
}


CCPU::~CCPU()
{
	for (
		map<string,IInstruction*>::iterator it1 =
		m_instruction.begin();
		it1 != m_instruction.end(); it1++)
	{
		delete it1->second;
	}

	for (
		vector<IInstruction*>::iterator it3 =
		m_program.begin();
		it3 != m_program.end(); it3++)
	{
		delete (*it3);
	}
}


void CCPU::addLabel(std::string const& rest)
{
	size_t colon = rest.find_first_of(':');

	if (colon == string::npos) throw EParse("No colon at end of label found");
	if (colon != rest.size()-1) throw EParse("residual characters after colon");

	if (rest.at(0) != '_' &&
	    rest.at(0) != '.') throw EParse("_ or . missing: at Label");

	string label = rest.substr(1, colon-1);
	if (label.find_first_of(' ') != string::npos) throw EParse ("No Space in Label allowed");
	if (label.find_first_of(',') != string::npos) throw EParse ("No Comma in Label allowed");
	if (m_label.find(label) != m_label.end()) throw EParse ("Label already defined");

	int where = m_asm_size+lowaddr;

	m_label[label] = where;
	cout << "adding label " << label << " with " <<
		where << " that is line " << offset(where)+1 << std::endl;
}

/**@param param is sthg. like "_display,256".
  We just need the information that display is there (number in it is
  incremented)*/
void CCPU::addArray(std::string const& param)
{
	if (param.at(0) != '_') throw EParse("No _ found before array");
	
	size_t sep = param.find_first_of(',');

	if (sep == string::npos) throw EParse("No sep at end of array found");
	if (sep == param.size()-1) throw EParse("No residual characters after sep");

	string label = param.substr(1, sep-2);
	int length = 1;

	m_array[label] = length;
}

void CCPU::loadProgram(istream &in)
{
	using namespace boost::algorithm;

	string line;
	size_t linenr=0;
	while (getline(in, line))
	{ try {
		++linenr;

		string cmd;
		stringstream ss (line);
		ss >> cmd;

		// Jump Empty lines
		if (cmd.size() == 0) continue;

		// Jump Comments
		if (cmd.at(0) == ';') continue;

		string rest;
		getline (ss, rest);
		trim(rest);

		// Handle Labels
		if (cmd.at(cmd.length()-1) == ':')
		{
			addLabel(cmd);
			continue;
		}

		// Handle Arrays
		if (cmd == ".comm")
		{
			addArray(rest);
			continue;
		}

		// Handle Instructions
		if (m_instruction.find(cmd) == m_instruction.end())
		{
			throw EParse("Unknown instruction found");
		}

		char_separator<char> sep(",");
		tokenizer<char_separator<char> > tokens(rest, sep);

		m_program.push_back(m_instruction[cmd]->clone());

		size_t i=0;
		for (
			tokenizer<char_separator<char> >::iterator it = tokens.begin();
			it != tokens.end();
			++it)
		{
			string t = (*it);
			trim (t);
			// if (t.find_first_of(' ') != string::npos) throw EParse ("No Space in Parameter allowed");
			m_program.back()->param(i, t);
			i++;
		}
		if (i<m_program.back()->size()) throw EParse ("Not enough arguments");

		cout << "added " << cmd << " having size: " << m_program.back()->asm_size() << std::endl;
		m_program.back()->where(m_asm_size); // give the instruction the information where it is
		m_asm_size += m_program.back()->asm_size(); // update the size of the program up to now
	} catch (EParse &e)
	{
		e.linenr = linenr;
		throw;
	}
	}
}

void CCPU::saveProgram(ostream & out)
{
	bool spacer_inserted = false;
	size_t size_at_begin = m_program.size();
	size_t size_during_writing = 0;

	for (size_t i=0;i<size_at_begin; i++)
	{
		if( m_program[i]->isDataSection())
		{
			// we have a data section, so insert spacer
			if (!spacer_inserted)
			{
				cout << "size is : " << m_asm_size << " so we fill: " << numregisters-m_asm_size << std::endl;
				m_program.push_back(SSpace(numregisters-m_asm_size).clone());
				spacer_inserted = true;
				// to avoid multiple insertions of spacer
				// when having multiple sections
			}

			// bring contents of the data section to the end of m_program
			cout << "filling in an Empty inside data section" << std::endl;
			m_program.push_back(SEmpty().clone());
			swap (m_program[i], m_program.back());
		}
	}

	for (size_t i=0;i<m_program.size(); i++)
	{
		m_program[i]->save(out);
		size_during_writing += m_program[i]->asm_size();
		if (i<size_at_begin && size_during_writing != counter::get())
		{
			stringstream str;
			str << "size mismatch at size " << size_during_writing
			    << ", instruction " << i << " should have "
			    << counter::get();
			throw EParse (str.str());
		}
	}
}
