#include "instruction.h"
#include "except.h"

#include <iostream>

#include <boost/lexical_cast.hpp>

using namespace std;


void revout (std::ostream &os, word_t current)
{
	word_t code;
	int iA, iB;

	rev (current, code, iA, iB);
	os << code << " " << iB << " " << iA << finish;
}

int parseNumber (std::string const& param, std::string const& errormsg)
{
	using boost::lexical_cast;
	using boost::bad_lexical_cast;
	try {
		return lexical_cast<int> (param);
	} catch (bad_lexical_cast &) {
		throw EParse(errormsg);
	}
}

std::string toString (int param, std::string const& errormsg)
{
	using boost::lexical_cast;
	using boost::bad_lexical_cast;
	try {
		return lexical_cast<std::string> (param);
	} catch (bad_lexical_cast &) {
		throw EParse(errormsg);
	}
}

#include "counter.h"

std::ostream& finish(std::ostream& os)
{
	counter::inc();
	return os << "\n";
}

IInstruction::~IInstruction()
{}

void CRegInstruction::param (size_t which, string const& param)
{
	// because of displacement
	if (!m_displacement && which >= 2) throw EParse("number of parameter mismatch");
	if ( m_displacement && which >= 3) throw EParse("number of parameter mismatch (with displacement)");

	int start = 0;
	int length = param.length();

	if (param.at(length-1) == '+')
	{
		m_info[which].postinc = true;
		length --;
	}
	else if (param.at(length-1) == '-')
	{
		m_info[which].postdec = true;
		length --;
	}

	if (param.at(start) == '@')
	{
		m_info[which].addr = true;
		start ++;
		length --;
	}

	if (param.at(start) == '(')
	{
		// displacement start

		start++;
		length--;

		try {
			int nr = parseNumber(param.substr(start,length), "Displacement must be a number or var (starts with _)");
			if (nr>=0) m_positive = true;
			else m_positive = false;
		} catch (EParse const& e)
		{
			if (param.at(start) != '_') throw;

			start ++;
			length --;
		}
		if (param.length() < 3) throw EParse ("first displacement not long enough");

		// we have a displacement, store that information
		m_displacement = which + 1;
	}

	if (param.at(length-1) == ')')
	{
		if (param.length() < 1) throw EParse ("second displacement not long enough");
		// displacement end
		string info = param.substr(0,param.length()-1);

		// cut of )
		length --;
	}

	if (param.at(start) == '+')
	{
		m_info[which].preinc = true;
		start ++;
		length --;
	}
	else if (param.at(start) == '-')
	{
		m_info[which].predec = true;
		start ++;
		length --;
	}

	if (param.at(start) == '_')
	{
		// @_display is not an address, take value as is
		m_info[which].addr = false;
		start ++;
		length --;

		if (param.at(start) == '_')
		{
			// remove __, because displacement code does that too
			start ++;
			length --;
		}
	}

	if (param.at(start) == '#')
	{
		if (m_info[which].addr) throw EParse ("Cant take address of immediate");
		m_info[which].immediate = true;
		start ++;
		length --;

		try {
			parseNumber(param.substr(start,length), "Immediate must be a number or var (starts with _)");
		} catch (EParse const& e)
		{
			if (param.at(start) != '_') throw;

			start ++;
			length --;

			if (param.at(start) == '_')
			{
				// remove __, because displacement code does that too
				start ++;
				length --;
			}
		}
	}

	m_param[which] = param.substr(start,length);

}

void CJumpInstruction::param (size_t which, string const& param)
{
	if (which >= size()) throw EParse("number of parameter mismatch");

	if (param.length() < 3) throw EParse("Function name too short");

	// a normal label
	if (param.at(0) == '.')
	{
		m_name = param.substr(1, param.length()-1);
		return;
	}

	// a function label
	if (param.at(0) != '@') throw EParse("@ missing: No call to a function");
	if (param.at(1) != '_') throw EParse("_ missing: Value not takes as is for a function");

	m_name = param.substr(2, param.length()-2);
}

void SSection::param (size_t which, string const& param)
{
	if (param == ".data") m_section = data;
	else if (param == ".rodata") m_section = rodata;
	else if (param == ".bss") m_section = bss;
	else if (param == ".text") m_section = text;
	else throw EParse("Section parsing failed");
}

void CAsmDirective::param (size_t which, string const& param)
{
	if (which >= size()) throw EParse("number of parameter mismatch");

	m_param = param;
}

/**Meta Directives don't need their parameters.

 It would be similar to array parameter*/
void CMetaDirective::param (size_t which, string const& param)
{}

void SEmpty::param (size_t which, std::string const& param)
{
	throw EParse ("Empty instruction does not take any parameter");
}

void SSpace::param (size_t which, std::string const& param)
{
	throw EParse ("Space instruction does not take any parameter");
}

void SAscii::param (size_t which, std::string const& param)
{
	if (which >= size()) throw EParse("number of parameter mismatch");

	size_t end = param.length() - 1;

	if (param[0] != '"')  throw EParse("no \" found in begin of ascii");
	if (param[end] != '"')  throw EParse("no \" found in end of ascii");
	if (param[end-1] != '0')  throw EParse("no \0 found in end of ascii");
	if (param[end-2] != '\\')  throw EParse("no \\ found in end of ascii");

	string tmp = param.substr(1,end-1);

	// spare out the last 2 characters "\0"
	for (size_t i =0; i<tmp.length()-2; i++)
	{
		if (tmp[i] == '\\')
		{
			// decode char?
		} else {
			m_string.push_back(tmp[i]);
		}
	}
}
