#ifndef INSTRUCTION_H
#define INSTRUCTION_H

#include <iostream>
#include <string>
#include <vector>
#include <map>

#include "except.h"

#include "config.h"

/** outputs data as triplet code iB iA.

  0 ist die niedrigste mögliche Addresse.
  8192 ist dann die Addresse 0.
*/
void revout (std::ostream &os, word_t current);

/** Parse and return a number param.*/
int parseNumber (std::string const& param, std::string const& errormsg);

std::ostream& finish(std::ostream& os);

class ICPU;

/**Interface for all instructions.*/
class IInstruction
{
protected:
	int address;
public:
	IInstruction() : address(0)
	{}
	/**Clone the instruction.*/
	virtual IInstruction* clone () = 0;
	/**Execute the instruction.
	 @pre this is a cloned instruction*/
	virtual void save (std::ostream &out) = 0;
	/**Number of arguments required.*/
	virtual size_t size () = 0;
	/**The size of the assembler output.
	  = number of lines in basic asm
	  = number of opcodes for simulator
	*/
	virtual size_t asm_size () = 0;
	virtual void param (size_t which, std::string const& what) = 0;

	/**Address is the address of the first assembler instruction
	  in this instruction.
	  +1 and so on is to refer to the subsequent opcodes within
	  one instruction.
	  */
	virtual void where (int address)
	{
		this->address = address + lowaddr;
	}

	/**returns true on SSection starting a data section
	  and on .ascii members.*/
	virtual bool isDataSection()
	{
		return false;
	}

	virtual ~IInstruction() = 0;
};

class CRegInstruction : public IInstruction
{
public:
	struct info
	{
		bool addr, postinc, postdec, preinc, predec, immediate;
	};

protected:
	std::vector<std::string> m_param;
	std::vector<info> m_info;
	int m_displacement; /// 0 for no displacement, 1 on the left side, 2 on the right side
	bool m_positive; /// true if the displacement is
	 /// * a positive number
	 /// * an array, which is above 8000
	 /// * the number 0
	 /// will be set to false otherwise

public:
	CRegInstruction() :
		m_param(3),
		m_info(3),
		m_displacement(false),
		m_positive(true)
	{}

	virtual void param (size_t which, std::string const& what);

	virtual ~CRegInstruction() {}
};


/**Moves from source to destination. (page 88)

  MOV.W Rs, Rd

  (R7 == SP)
  MOV.W Rs, @–R7 is identical in machine language to PUSH.W Rs.
  POP.W Rn is identical in machine language to MOV.W @SP+, Rn.

  This instruction moves one word of data from a source register to a destination register and
  sets condition code flags according to the data value.

  Moving of programcounter (pc) does not work!
  This operation is not atomar.
  use RTS and JSR instead.

  */
class CMov : public CRegInstruction
{
public:
	CMov* clone ()
	{
		return new CMov(*this);
	}
	void save(std::ostream &os)
	{
		if (m_displacement == 1)
		{
			// displacement on the left side
			os << "setc " << m_param[0] << " W0" << finish;
			os << "set " << m_param[1] << " W" << finish;
			if (m_positive) os << "add W0 W" << finish;
			else os << "sub W0 W" << finish;
			os << "adda W " << m_param[2] << finish;
		}
		else if (m_displacement == 2)
		{
			// displacement on the right side
			os << "setc " << m_param[1] << " W0" << finish;
			os << "set " << m_param[2] << " W" << finish;
			if (m_positive) os << "add W0 W" << finish;
			else os << "sub W0 W" << finish;
			os << "seta " << m_param[0] << " W" << finish;
		}
		else
		{
			if (m_info[0].preinc) os << "add one " << m_param[0] << finish;
			if (m_info[0].predec) os << "sub one " << m_param[0] << finish;

			if (m_info[1].preinc) os << "add one " << m_param[1] << finish;
			if (m_info[1].predec) os << "sub one " << m_param[1] << finish;

			if (m_info[0].immediate)
			{
				if (m_info[1].addr) os << "setac " << m_param[0] << " " << m_param[1] << finish;
				else os << "setc " << m_param[0] << " " << m_param[1] << finish;
			}
			else if (m_info[1].addr) os << "seta " << m_param[0] << " " << m_param[1] << finish;
			else if (m_info[0].addr) os << "adda " << m_param[0] << " " << m_param[1] << finish;
			else os << "set " << m_param[0] << " " << m_param[1] << finish;

			if (m_info[0].postinc) os << "add one " << m_param[0] << finish;
			if (m_info[0].postdec) os << "sub one " << m_param[0] << finish;

			if (m_info[1].postinc) os << "add one " << m_param[1] << finish;
			if (m_info[1].postdec) os << "sub one " << m_param[1] << finish;
		}
	}
	size_t asm_size()
	{
		if (m_displacement)
		{
			return 4;
		} else {
			return 1 +
				m_info[0].preinc + m_info[0].predec +
				m_info[1].preinc + m_info[1].predec +
				m_info[0].postinc + m_info[0].postdec +
				m_info[1].postinc + m_info[1].postdec;
		}
	}
	size_t size()
	{
		return 2;
	}
};


/**Add number.

 This instruction adds word data in two general registers and places
 the result in the second
 general register.

 ADD (ADD binary) (word)

 <Operation>
 Rd + Rs -> Rd

  */
class CAdd : public CRegInstruction
{
public:
	CAdd* clone ()
	{
		return new CAdd(*this);
	}
	void save(std::ostream &os)
	{
		os << "add " << m_param[0] << " " << m_param[1] << finish;
	}
	size_t asm_size()
	{
		return 1;
	}
	size_t size()
	{
		return 2;
	}
};

/**Add a constant number (1 or 2).

  This instruction adds the immediate value 1 or 2 to word data in a general register. Differing
  from the ADD instruction, it does not affect the condition code flags.

  In contrast to CSubConstant it could add any number.

  */
class CAddConstant : public CRegInstruction
{
public:
	CAddConstant* clone ()
	{
		return new CAddConstant(*this);
	}
	void save(std::ostream &os)
	{
		os << "addc " << m_param[0] << " " << m_param[1] << finish;
	}
	size_t asm_size()
	{
		return 1;
	}
	size_t size()
	{
		return 2;
	}
};

/** Subtract two registers. (SUB, page 117)

  This instruction subtracts a 16-bit source register from a 16-bit destination register and places
  the result in the destination register.

  SUB.W Rs, Rd

  */
class CSub : public CRegInstruction
{
public:
	CSub* clone ()
	{
		return new CSub(*this);
	}
	void save(std::ostream &os)
	{
		os << "sub " << m_param[0] << " " << m_param[1] << finish;
	}
	size_t asm_size()
	{
		return 1;
	}
	size_t size()
	{
		return 2;
	}
};

/**Sub a constant number (1 or 2).

  This instruction subtracts the immediate value 1 or 2 from word data in a general register.
  Differing from the SUB instruction, it does not affect the condition code flags.
  The SUBS instruction does not permit byte operands.


  */
class CSubConstant : public CRegInstruction
{
public:
	CSubConstant* clone ()
	{
		return new CSubConstant(*this);
	}
	void save(std::ostream &os)
	{
		if (m_param[0] == "1") os << "sub one " << m_param[1] << finish;
		else os << "sub two " << m_param[1] << finish;
	}
	size_t asm_size()
	{
		return 1;
	}
	size_t size()
	{
		return 2;
	}
};



/**Compare two words. (CMP, page 73)

  This instruction subtracts a source register from a destination register and sets the condition
  code flags according to the result. The destination register is not altered.

  Z will be changed.
  Z is containing 0 if they are equal.
  something else otherwise.

  */
class CCompare : public CRegInstruction
{
public:
	CCompare* clone ()
	{
		return new CCompare(*this);
	}
	void save(std::ostream &os)
	{
		os << "set " << m_param[0]  << " Z" << finish;
		os << "sub " << m_param[1] << " Z" << finish;
	}
	size_t asm_size()
	{
		return 2;
	}
	size_t size()
	{
		return 2;
	}
};



/**Return from Subroutine. (page 108, RTS)

  POP.W Rn is identical in machine language to MOV.W @SP+, Rn.

  This instruction returns from a subroutine. It pops the program
  counter (PC) from the stack.
  Program execution continues from the address restored to the program
  counter.
  The PC contents at the time of execution of this instruction are lost.
*/
class CReturn : public IInstruction
{
public:
	CReturn* clone ()
	{
		return new CReturn(*this);
	}
	void save(std::ostream &os)
	{
		os << "adda SP W" << finish; // pop program pointer from stack!
		os << "add one SP" << finish; // increment program pointer
		os << "set W PC" << finish; // next instruction is where we came from
	}
	size_t asm_size()
	{
		return 3;
	}
	size_t size()
	{
		return 0;
	}
	void param (size_t which, std::string const& what)
	{
		throw EParse ("CReturn: rts does not take a parameter");
	}
};

/** Top of all Jump instructions*/
class CJumpInstruction : public IInstruction
{
protected:
	std::string m_name;
public:
	CJumpInstruction() : m_name()
	{}
	virtual void param (size_t which, std::string const& what);
	virtual size_t size() {return 1;}
};


/** Always jump.
 */
class CJump: public CJumpInstruction
{
	const std::map <std::string, int> &m_label;
public:
	CJump (std::map <std::string, int> const& label) :
		m_label(label)
	{}
	CJump* clone ()
	{
		return new CJump(*this);
	}
	void save(std::ostream &os)
	{
		std::map <std::string, int>::const_iterator iterator = m_label.find(m_name);

		if (iterator == m_label.end()) throw EParse(std::string("Cant find label ") + m_name);
		int m_pc = iterator->second;
		os << "jump " << m_pc << " one" << finish;
	}
	size_t asm_size()
	{
		return 1;
	}
	size_t size()
	{
		return 1;
	}
};


/** Jump if equal.
 */
class CJumpEqual: public CJumpInstruction
{
	const std::map <std::string, int> &m_label;
public:
	CJumpEqual (std::map <std::string, int> const& label) :
		m_label(label)
	{}
	CJumpEqual* clone ()
	{
		return new CJumpEqual(*this);
	}
	void save(std::ostream &os)
	{
		std::map <std::string, int>::const_iterator iterator = m_label.find(m_name);

		if (iterator == m_label.end()) throw EParse(std::string("Cant find label ") + m_name);
		int m_pc = iterator->second;
		os << "jump " << address+3 << " Z" << finish; // jump if Z != 0 (e.g. 5) to @HERE
		os << "setc 1 Z" << finish; // we did not jump, so it was 0, set it to 1 now
		os << "addc 2 PC" << finish; // jump over next line
		os << "setc 0 Z" << finish; // we jumped here, it was not zero(e.g. 5), change it to 0 <- HERE
		os << "jump " << m_pc << " Z" << finish;
	}
	size_t asm_size()
	{
		return 5;
	}
	size_t size()
	{
		return 1;
	}
};

/** Jump if not equal.
 */
class CJumpNotEqual: public CJumpInstruction
{
	const std::map <std::string, int> &m_label;
public:
	CJumpNotEqual (std::map <std::string, int> const& label) :
		m_label(label)
	{}
	CJumpNotEqual* clone ()
	{
		return new CJumpNotEqual(*this);
	}
	void save(std::ostream &os)
	{
		std::map <std::string, int>::const_iterator iterator = m_label.find(m_name);

		if (iterator == m_label.end()) throw EParse(std::string("Cant find label ") + m_name);
		int m_pc = iterator->second;
		os << "jump " << m_pc << " Z" << finish;
	}
	size_t asm_size()
	{
		return 1;
	}
	size_t size()
	{
		return 1;
	}
};


/** Jump to subroutine. (Page 85, JSR)

 This instruction pushes the program counter onto the stack, then
 branches to a specified
 destination address. The program counter value pushed on the stack is
 the address of the
 instruction following the JSR instruction. The destination address must
 be even.

 PUSH Rn is equivalent to MOV.W Rn, @-SP.

 <Operation>
 PC -> @-SP
 (EAd) -> PC

 */
class CJumpToSubroutine : public CJumpInstruction
{
	const std::map <std::string, int> &m_label;
public:
	CJumpToSubroutine (std::map <std::string, int> const& label) :
		m_label(label)
	{}
	CJumpToSubroutine* clone ()
	{
		return new CJumpToSubroutine(*this);
	}
	void save(std::ostream &os)
	{
		if (m_name == "__key")
		{
			os << "setc -1 R0" << finish; // set return value to -1
			os << "key " << address + 3 << " W" << finish; // jumps to last line having input, stores value to W
			os << "addc 2 PC" << finish; // may be skipped, will skip next line
			os << "set W R0" << finish; // will return value
		}
		else if (m_name == "__strncmp")
		{
#include "strncmp.cpp"
		}
		else if (m_name == "__get")
		{
			os << "adda R0 R0" << finish;
		}
		else if (m_name == "__set")
		{
			os << "seta R1 R0" << finish;
		}
		else if (m_name == "__read")
		{
			os << "read flash R0" << finish;
		}
		else if (m_name == "__write")
		{
			os << "write flash R0" << finish;
		}
		else if (m_name == "__modhi3")
		{
			os << "mod R1 R0" << finish;
		}
		else if (m_name == "__mulhi3")
		{
			os << "mul R1 R0" << finish;
		}
		else if (m_name == "__divhi3")
		{
			os << "div R1 R0" << finish;
		} else {
			std::map <std::string, int>::const_iterator iterator = m_label.find(m_name);

			if (iterator == m_label.end()) throw EParse(std::string("Cant find label ") + m_name);
			int m_pc = iterator->second;
			os << "sub one SP" << finish; // decrement program pointer
			os << "set PC W" << finish; // save program counter
			os << "addc 4 W" << finish; // correct return address
			os << "seta W SP" << finish; // push program pointer to stack!
			os << "jump " << m_pc << " one" << finish;
		}
	}
	size_t asm_size()
	{
		if (m_name == "__key") return 4;
		else if (m_name == "__strncmp") return 53;
		else if (m_name == "__get") return 1;
		else if (m_name == "__set") return 1;
		else if (m_name == "__read") return 1;
		else if (m_name == "__write") return 1;
		else if (m_name == "__modhi3") return 1;
		else if (m_name == "__mulhi3") return 1;
		else if (m_name == "__divhi3") return 1;
		else return 5;
	}
	size_t size()
	{
		return 1;
	}
};



class CAsmDirective : public IInstruction
{
protected:
	std::string m_param;
public:
	virtual void param (size_t which, std::string const& what);
	CAsmDirective() : m_param()
	{}

	virtual ~CAsmDirective() {}
};

/**Simply jump to the subroutine main.*/
class SSection : public CJumpToSubroutine
{
	enum ESection
	{
		none, /// no section was deduced up to now
		data, /// initialized data which have a defined value in the beginning
		rodata, /// as above but read only
		bss, /// data initialized with 0 (block started by symbol)
		text /// the program
	} m_section;
public:
	virtual void param (size_t which, std::string const& what);

	SSection(std::map <std::string, int> const& label) :
		CJumpToSubroutine (label),
		m_section(none)
	{
		m_name = "main";
	}
	SSection* clone ()
	{
		return new SSection(*this);
	}
	void save(std::ostream &os)
	{
		if (m_section == text)
		{
			// pre main code here
			CJumpToSubroutine::save(os); // call the main function
			// exit the program
			os << "key " << lowaddr << " " << lowaddr << finish;
		}
	}
	size_t asm_size()
	{
		if (m_section == text) return CJumpToSubroutine::asm_size()+1;
		else return 0;
	}
	size_t size()
	{
		return 1;
	}
	ESection section()
	{
		return m_section;
	}
	bool isDataSection()
	{
		return m_section == data || m_section == rodata;
	}
};


class SAlign : public CAsmDirective
{
public:
	SAlign* clone ()
	{
		return new SAlign(*this);
	}
	void save(std::ostream &os)
	{
	}
	size_t asm_size()
	{
		return 0;
	}
	size_t size()
	{
		return 1;
	}
};


class SGlobal : public CAsmDirective
{
public:
	SGlobal* clone ()
	{
		return new SGlobal(*this);
	}
	void save(std::ostream &os)
	{
	}
	size_t asm_size()
	{
		return 0;
	}
	size_t size()
	{
		return 1;
	}
};


class SEnd : public CAsmDirective
{
public:
	SEnd* clone ()
	{
		return new SEnd(*this);
	}
	void save(std::ostream &os)
	{
	}
	size_t asm_size()
	{
		return 0;
	}
	size_t size()
	{
		return 0;
	}
};

/**Sreates an empty instruction.
 It does not output anything, it can be filled in
 anywhere safely.*/
class SEmpty: public IInstruction
{
public:
	SEmpty()
	{}

	virtual SEmpty* clone ()
	{
		return new SEmpty(*this);
	}

	virtual void save (std::ostream &out)
	{}

	virtual size_t size ()
	{
		return 0;
	}

	virtual size_t asm_size ()
	{
		return 0;
	}

	virtual void param (size_t which, std::string const& what);

	virtual ~SEmpty()
	{}
};

/**Outputs n Zeros to fill up space.
  */
class SSpace: public IInstruction
{
	size_t m_space;
public:
	SSpace(size_t space) :
		m_space(space)
	{}

	virtual SSpace* clone ()
	{
		return new SSpace(*this);
	}

	virtual void save (std::ostream &out)
	{
		for (size_t i=0; i<m_space; ++i)
		{
			revout(out, 0);
		}
	}

	virtual size_t size ()
	{
		return 0;
	}

	virtual size_t asm_size ()
	{
		return m_space;
	}

	virtual void param (size_t which, std::string const& what);

	virtual ~SSpace()
	{}
};

/**Outputs an ascii array to flash storage.

  So the caller must make sure that the instruction is started
  at the beginning of a flash storage unit.
  */
class SAscii: public IInstruction
{
	std::string m_string;
public:
	SAscii() :
		m_string()
	{}

	virtual SAscii* clone ()
	{
		return new SAscii(*this);
	}

	virtual void save (std::ostream &out)
	{
		size_t i;
		for (i=0; i<m_string.length(); ++i)
		{
			revout(out, m_string[i]);
		}

		for (;i<128; i++)
		{
			revout(out, 0);
		}
	}
	virtual size_t size ()
	{
		return 1;
	}
	/**Always return 0.
	 asm_size is always 0 to not confuse other instructions.
	 But in fact it is always 128 registers (=512 bytes) long.
	*/
	virtual size_t asm_size ()
	{
		return 0;
	}
	virtual bool isDataSection()
	{
		return true;
	}

	virtual void param (size_t which, std::string const& what);

	virtual ~SAscii()
	{}
};


class CMetaDirective : public IInstruction
{
protected:
	std::string m_param;
public:
	virtual void param (size_t which, std::string const& what);
	CMetaDirective() : m_param()
	{}

	virtual ~CMetaDirective() {}
};

class SFile : public CMetaDirective
{
public:
	SFile* clone ()
	{
		return new SFile(*this);
	}
	void save(std::ostream &os)
	{
		// 7 basic registers
		revout (os, start); // pc
		revout (os, -1); // neg
		revout (os, 0); // zero
		revout (os, 1); // one
		revout (os, 2); // two
		revout (os, 0); // Z zero cmp register
		revout (os, 0); // W

		// 8 working registers
		revout (os, 0); // W0
		revout (os, 0); // W1
		revout (os, 0); // W2
		revout (os, 0); // W3
		revout (os, 0); // W4
		revout (os, 0); // W5
		revout (os, 0); // W6
		revout (os, 0); // W7

		// 2 not used
		revout (os, 100); // N0
		revout (os, 200); // N1

		// 8 registers
		revout (os, 0); // R0
		revout (os, 0); // R1
		revout (os, 0); // R2
		revout (os, 0); // R3
		revout (os, 0); // R4
		revout (os, 0); // R5
		revout (os, 0); // R6
		revout (os, 8063); // R7 (=SP)
	}
	size_t asm_size()
	{
		return 25;
	}
	size_t size()
	{
		return 1;
	}
};

class SIdent : public CMetaDirective
{
public:
	SIdent* clone ()
	{
		return new SIdent(*this);
	}
	void save(std::ostream &os)
	{
	}
	size_t asm_size()
	{
		return 0;
	}
	size_t size()
	{
		return 1;
	}
};


#endif
