#include "cpu.h"

#include "register.h"
#include "flash.h"
#include "display.h"
#include "keyboard.h"

#include "error.h"

#include <stdexcept>
#include <iostream>

#include "getopt_pp.h"

using namespace GetOpt;
using namespace std;

void help();
void test(string const&);

int main(int argc, char ** argv) try
{
	string file;
	bool needs_help, needs_test;
	GetOpt_pp options (argc, argv);
	options.exceptions(ios::failbit | ios::eofbit);

	options >> Option('f', "file", file, "flash.img")
		>> OptionPresent('t', "test", needs_test)
		>> OptionPresent('h', "help", needs_help);

	if (needs_help)
	{
		help();
		return 0;
	}

	if (needs_test)
	{
		test(file);
		return 0;
	}

	cout << "cpu is simulating with flash " << file << endl;

	Register reg;
	Flash flash (file);
	Display dis (reg.display());
	Keyboard key;

	flash.writeRegister(reg);

	CPU cpu (reg, flash, dis, key);

	while (true)
	{
		cpu.exec();
	}

} catch(GetOptEx const& e)
{
	cerr << "Invalid Options" << endl;
	help();
	return 1;
} catch (out_of_range const& oor)
{
	cerr << oor.what() << endl;
	return 2;
} catch (file_error const& fe)
{
	cerr << "Could not open flash file: " << fe.what() << endl;
	help();
	return 3;
} catch (ifstream::failure const& e)
{
	cerr << "Exception reading file: " << e.what() << endl;
} catch (zero_division_error const& zde)
{
	cerr << "Processor tried to divide through zero" << endl;
	return 4;
} catch (exit const& e)
{
	cout << "Exit with returncode " << e.ret << endl;
	return e.ret;
}catch (...)
{
	cerr << "Caught unknown exception" << endl;
	return 100;
}

void help()
{
	cout << "Hilfe zu Simulator" << endl;
	cout << endl;
	cout << "Geben sie mit -f ein flash Speicher an." << endl;
	cout << "Dieser muss zuerst mit einem Assembler" << endl;
	cout << "erzeugt werden" << endl;
	cout << endl;
	cout << "In der Flashdatei werden Opcodes im Format oBA" << endl;
	cout << "ausgeführt, die je 32bit gross sind." << endl;
	cout << "Informationen wie diese interpretiert werden," << endl;
	cout << "siehe opcode.cpp." << endl;
	cout << endl;
	cout << " -t   Führe ein paar interne Tests durch," << endl;
	cout << "      anstatt den Simulator auszuführen." << endl;
}

void test(string const& file)
{
	Flash flash (file);
	Register reg;

	flash.writeRegister(reg);

	reg[-8192] = 'a';
	reg[-8191] = 'b';
	reg[lowflash] = 'u';
	reg[lowflash+1] = 'v';
	reg[highflash-1] = 'k';
	reg[highflash] = 'l';
	reg[0] = 'x';
	reg[1] = 'y';
	flash.writeBlock(reg, 20, lowflash);

	// cleanup
	reg[lowflash] = ' ';
	reg[lowflash+1] = ' ';
	reg[highflash-1] = ' ';
	reg[highflash] = ' ';

	flash.readBlock(reg, 20, lowflash);
	cout << (char) reg[-8192] << endl;
	cout << (char) reg[-8191] << endl;
	cout << (char) reg[lowflash] << endl;
	cout << (char) reg[lowflash+1] << endl;
	cout << (char) reg[highflash-1] << endl;
	cout << (char) reg[highflash] << endl;
	cout << (char) reg[0] << endl;
	cout << (char) reg[1] << endl;

	for (int i=lowflash-1; i< highflash+1; i++) cout << (char) reg[i] << '_';
	cout << endl;
}
