#include "opcode.h"

#include "register.h"

#include "error.h"

#ifdef DEBUG
#include <iostream>
#include <iomanip>
#endif


/** setc (0): schreibe Konstante.

  Set B (as constant) in register A.
  (Jump to Address B for A = pc)
  */
void SetConstantOpcode::operator() (int A, int B)
{
	reg[A] = B;
	if (A!=pc) reg[pc]++;
}

/**set (1): schreibe (kopiere Register)

  Write the content of register B into register A.
  (Indirect jump for A = pc)
  */
void SetOpcode::operator() (int A, int B)
{
	reg[A] = reg[B];
	if (A!=pc) reg[pc]++;
}

/**setac (2): indirektes Schreiben v. Konstante
  
  Write B (as constant) in the register, which address
  is the content from register A.
  (Indirect jump if content from A = pc)
  */
void SetAddressConstantOpcode::operator() (int A, int B)
{
	reg[reg[A]] = B;
	if (reg[A] != pc) reg[pc]++;
}

/**seta (3): indirektes Schreiben (kopieren) von Register.
  
  Write the content of register B in the register,
  which address is the content from register A.
  */
void SetAddressOpcode::operator() (int A, int B)
{
	reg[reg[A]] = reg[B];
	if (reg[A] != pc) reg[pc]++;
}

/**addc (4): addiere zu Register eine Konstante.
  
  Add B (as constant) to the content of register A
  and write the content to register A.
  (relative Jump for A = pc)
 */
void AddConstantOpcode::operator() (int A, int B)
{
	reg[A] += B;
	if (A!=pc) reg[pc]++;
}

/**add (5): addiere zu Register ein anderes Register.
  
  Add the content of register B in the register A.
  (relative Indirect Jump for A = pc)
  */
void AddOpcode::operator() (int A, int B)
{
	reg[A] += reg[B];
	if (A!=pc) reg[pc]++;
}

/**adda (6): lade vom indirekten Register.

  AKTUELL: Aufgabenstellung

  Diese Methode dient zum indirekten laden!

  Nimm den Wert des durch die Adresse im Register B stehenden
  Register und schreibe es in das Register A.


  Load the content of register A from the register,
  which address is the content from register B.

  Can be used as Indirect Loading Operation.
  */
void AddAddressOpcode::operator() (int A, int B)
{
	reg[A] = reg[reg[B]];
	if (A!= pc) reg[pc]++;
}

/**sub (7): subtrahiere Register.

  Subtract the content of register B - A and
  write it to register A.
  */
void SubOpcode::operator() (int A, int B)
{
	reg[A] -= reg[B];
	if (A!=pc) reg[pc]++;
}

/**mulc (8): Multipliziere Konstante zu register
  
  Multiply B (as constant) to the content of register A
  and write the content to register A.
  (relative Jump for A = pc)
 */
void MulConstantOpcode::operator() (int A, int B)
{
	reg[A] *= B;
	if (A!=pc) reg[pc]++;
}

/**mul(9): Multipliziere Register zu Register.
  
  Multiply the content of register B in the register A.
  (relative Indirect Jump for A = pc)
  */
void MulOpcode::operator() (int A, int B)
{
	reg[A] *= reg[B];
	if (A!=pc) reg[pc]++;
}

/**mod(10=a): Dividiere Konstante.
  
  The content of the register B modulo the content of register A
  and write the content to register A.
  (relative Jump for A = pc)
 */
void ModOpcode::operator() (int A, int B)
{
	if (reg[B]==0) throw zero_division_error();
	reg[A] %= reg[B];
	if (A!=pc) reg[pc]++;
}

/**div(11=b): Dividiere Register zu anderem Register.
  
  Divide the content of register B in the register A.
  (relative Indirect Jump for A = pc)
  */
void DivOpcode::operator() (int A, int B)
{
	if (reg[B]==0) throw zero_division_error();
	reg[A] /= reg[B];
	if (A!=pc) reg[pc]++;
}

#include "keyboard.h"
#include "display.h"

/**jump(12=c): Springe wenn ungleich 0.
  
  Jump to address B (writing B as constant into pc)
  if register A has a value != zero, otherwise only
  increment pc.*/
void JumpOpcode::operator() (int A, int B)
{
#ifdef DEBUG
	using namespace std;
	cout << "comparing reg " << dec << A 
		<< " which is " << reg[A] << endl;
#endif
	if (reg[A]!=0)
	{
#ifdef DEBUG
		cout << "jump to: " << dec << B << " which is line: " << offset(B)+1 << endl;
#endif
		reg[pc] = B;
	} else {
		reg[pc]++;
	}
}

/**key (13=d): Lese ein Zeichen von Tastatur und springe
               wenn etwas gelesen.
  
  Read a character from the keyboard,
  and save it into register A.
  Jump to address B (B is written to pc)
  when a character could be successfully read,
  otherwise just increment pc.
  Display is shown before.

  B is a constant!
  */
void KeyOpcode::operator() (int A, int B)
{
	if (A == lowaddr && B == lowaddr)
	{
		throw exit(reg[R]);
	}

	display.show();
	int code = key.read();

#ifdef DEBUG
	using namespace std;
	cout << "code: " << dec << code
		<< " A: " << A
		<< " B: " << B
		<< endl;

#endif


	if (code)
	{
#ifdef DEBUG
		cout << "jump to: " << dec << B << " which is line: " << offset(B)+1 << endl;
		cout << "save: " << dec << code << " to " << A << endl;
#endif
		reg[A] = code;
		reg[pc] = B;
	} else {
		reg[pc]++;
	}
}

#include "flash.h"

/**write (14=e): schreibe flash.
  
  Write the content of 128 successive
  register into the flash.
  B is the constant address of the register
  where to read from.
  The content from A is a flash address
  to write to.
 */
void WriteOpcode::operator() (int A, int B)
{
	flash.writeBlock(reg, reg[A], B);
	reg[pc]++;
}

/**read (15=f): lese flash.
  
  Read 128 successive bytes from flash
  and write it into register.
  B is the constant address of the register
  where to write to.
  The content from A is the flash address
  to read from.
  */
void ReadOpcode::operator() (int A, int B)
{
	flash.readBlock(reg, reg[A], B);
	reg[pc]++;
}
