#ifndef OPCODE_H
#define OPCODE_H

#include "config.h"
#include "error.h"

/**Opcode ist eine ausführbare Einheit.
  Das Register, Flash oder Display wird direkt
  bearbeitet.

  Es handelt sich hier um ein Funktor, also ein
  Funktionsobjekt (ein Objekt welches wie sich
  wie eine Funktion verhalten kann). Das Ausführen
  der Funktion entspricht der Funktion des Opcodes.
  Subklassen übernehmen die tatsächliche Implementierung.

  Die benötigten Referenzen wie Register, Flash, Display,
  Keyboard usw. werden den Subklassen bei der Erzeugung
  angegeben.
 */
class Opcode
{
public:
	virtual void operator() (int A, int B)=0;
	virtual const char* name() const=0;
	virtual ~Opcode() {}
};

class Register;

class SetConstantOpcode : public Opcode
{
	Register &reg;
public:
	SetConstantOpcode (Register &reg): reg(reg)
	{}
	virtual void operator() (int A, int B);
	const char* name() const {return "setc";}
};

class SetOpcode : public Opcode
{
	Register &reg;
public:
	SetOpcode (Register &reg): reg(reg)
	{}
	virtual void operator() (int A, int B);
	const char* name() const {return "set";}
};

class SetAddressOpcode : public Opcode
{
	Register &reg;
public:
	SetAddressOpcode (Register &reg): reg(reg)
	{}
	virtual void operator() (int A, int B);
	const char* name() const {return "seta";}
};

class SetAddressConstantOpcode : public Opcode
{
	Register &reg;
public:
	SetAddressConstantOpcode (Register &reg): reg(reg)
	{}
	virtual void operator() (int A, int B);
	const char* name() const {return "setac";}
};

class AddConstantOpcode : public Opcode
{
	Register &reg;
public:
	AddConstantOpcode (Register &reg): reg(reg)
	{}
	virtual void operator() (int A, int B);
	const char* name() const {return "addc";}
};

class AddOpcode : public Opcode
{
	Register &reg;
public:
	AddOpcode (Register &reg): reg(reg)
	{}
	virtual void operator() (int A, int B);
	const char* name() const {return "add";}
};

class AddAddressOpcode : public Opcode
{
	Register &reg;
public:
	AddAddressOpcode (Register &reg): reg(reg)
	{}
	virtual void operator() (int A, int B);
	const char* name() const {return "adda";}
};

class SubOpcode : public Opcode
{
	Register &reg;
public:
	SubOpcode (Register &reg): reg(reg)
	{}
	virtual void operator() (int A, int B);
	const char* name() const {return "sub";}
};

class MulConstantOpcode : public Opcode
{
	Register &reg;
public:
	MulConstantOpcode (Register &reg): reg(reg)
	{}
	virtual void operator() (int A, int B);
	const char* name() const {return "mulc";}
};

class MulOpcode : public Opcode
{
	Register &reg;
public:
	MulOpcode (Register &reg): reg(reg)
	{}
	virtual void operator() (int A, int B);
	const char* name() const {return "mul";}
};

class ModOpcode : public Opcode
{
	Register &reg;
public:
	ModOpcode (Register &reg): reg(reg)
	{}
	virtual void operator() (int A, int B);
	const char* name() const {return "divc";}
};

class DivOpcode : public Opcode
{
	Register &reg;
public:
	DivOpcode (Register &reg): reg(reg)
	{}
	virtual void operator() (int A, int B);
	const char* name() const {return "div";}
};

class JumpOpcode : public Opcode
{
	Register &reg;
public:
	JumpOpcode (Register &reg): reg(reg)
	{}
	virtual void operator() (int A, int B);
	const char* name() const {return "jump";}
};

class Keyboard;
class Display;

class KeyOpcode : public Opcode
{
	Register &reg;
	Keyboard &key;
	Display &display;
public:
	KeyOpcode (Register &reg, Keyboard &key, Display &display):
		reg(reg),
		key(key),
		display(display)
	{}
	virtual void operator() (int A, int B);
	const char* name() const {return "key";}
};

class Flash;

class WriteOpcode : public Opcode
{
	Register &reg;
	Flash &flash;
public:
	WriteOpcode (Register &reg, Flash &flash):
		reg(reg),
		flash(flash)
	{}
	virtual void operator() (int A, int B);
	const char* name() const {return "write";}
};

class ReadOpcode : public Opcode
{
	Register &reg;
	Flash &flash;
public:
	ReadOpcode (Register &reg, Flash &flash):
		reg(reg),
		flash(flash)
	{}
	virtual void operator() (int A, int B);
	const char* name() const {return "read";}
};

#endif
